{+--------------------------------------------------------------------------+
 | Unit:        mwSor
 | Created:     12.97 -9.98
 | Author:      Martin Waldenburg
 | Copyright    1997 - 1998 all rights reserved.
 | Description: TmSor is a study how to create a powerful sorter
 |              with a minimum of work.
 |              Fixed records only.
 | Version:     1.6
 | Status       FreeWare
 | It's provided as is, without a warranty of any kind.
 | You use it at your own risc.
 | E-Mail me at Martin.Waldenburg@t-online.de
 +--------------------------------------------------------------------------+}


unit mwSor;

interface

uses
  Windows,
  SysUtils,
  Classes;

type
  TmSorCompare = function(Item1, Item2: Pointer): Integer;

type
  TmList = class(TList)
  private
    FDelta: Integer;
  protected
  public
    constructor Create;
    procedure Grow; override;
    property Delta: Integer read FDelta write FDelta;
  published
  end;  { TmList }

Type
  TmSorIO = class(TObject)
  private
    IOStream: TFileStream;
    fFilledSize:Longint;
    fBufferSize: LongInt;
    fBufferPos: LongInt;
    fBuffer: Pointer;
    fNeedFill: Boolean;
    fEof:Boolean;
    fFileEof: Boolean;
    FRecCount: Cardinal;
    fSize:Longint;
    fFilePos: LongInt;
    fDataLen:Longint;
    procedure AllocBuffer(NewValue:Longint);
  protected
  public
    constructor create(Stream: TFileStream; DataLen, BuffSize: Integer);
    destructor destroy;override;
    procedure FillBuffer;
    function ReadData:Pointer;
    procedure WriteData(Var NewData);
    procedure FlushBuffer;
    property Eof:Boolean read fEof;
    property RecCount: Cardinal read FRecCount;
    property Size:Longint read fSize;
    property DataLen:Longint read fDataLen;
    property FilePos:Longint read fFilePos;
  published
  end;  { TmSorIO }

 Type
  TmMergePart = class(TObject)
  private
    fPartStream:TFileStream;
    PartFilePos: LongInt;
    RecsToRead: LongInt;
    RecsReaded: LongInt;
    fBufferSize: LongInt;
    fBufferPos: LongInt;
    fBuffer: Pointer;
    fNeedFill: Boolean;
    fEof:Boolean;
    FRecCount: Cardinal;
    fDataLen:Longint;
    fData:Pointer;
    fNumber:Integer;
    procedure AllocBuffer(NewValue:Longint);
    procedure FillBuffer;
  protected
  public
    constructor create(Stream: TFileStream; FilePos, DataLen, Count, aNumber: LongInt);
    destructor destroy;override;
    procedure next;
    procedure Init;
    property Eof:Boolean read fEof;
    property Data:Pointer read fData;
    property Number:Integer read fNumber;
  published
  end;  { TmMergePart }

type
  TmwSor = class(TObject)
  private
    FParts: TmList;
    ReadStream: TFileStream;
    MergeStream: TFileStream;
    WriteStream: TFileStream;
    Reader: TmSorIo;
    Writer: TmSorIo;
    FRecordLen:Integer;
    SorList: TmList;
    SorFileName: String;
    TempFileName: String;
    procedure InitMerge;
    procedure Merge;
    procedure LooserSort;
    procedure CalculateBuffers;
  protected
  public
    constructor Create(RecLen: Integer);
    destructor Destroy; override;
    procedure Start(InFile, OutFile: String;  Compare: TmSorCompare);
  end;  { TmwSor }

var
  SorCompare: TmSorCompare;
  ReadBuffSize: LongInt;
  WriteBuffSize: LongInt;
  PartBuffSize: LongInt;

implementation

function StableCompare(P1, P2: Pointer): Integer;
begin
  Result:= SorCompare(TmMergePart(P1).Data, TmMergePart(P2).Data);
  if Result = 0 then
  begin
    if TmMergePart(P1).Number < TmMergePart(P2).Number then Result:= -1;
    if TmMergePart(P1).Number > TmMergePart(P2).Number then Result:= 1;
  end;
end; { StableCompare }

function MakeQuickSortStable(P1, P2: Pointer): Integer;
begin
  Result:= SorCompare(P1, P2);
  if Result = 0 then
  begin
    if Integer(P1) < Integer(P2) then Result:= -1;
    if Integer(P1) > Integer(P2) then Result:= 1;
  end;
end; { StableCompare }

constructor TmList.Create;
begin
    inherited create;
   FDelta:= 1024;
end;  { Create }

procedure TmList.Grow;
begin
  SetCapacity(Capacity + FDelta);
end;   { Grow }


constructor TmSorIO.create(Stream: TFileStream; DataLen, BuffSize: Integer);
begin
  IOStream:= Stream;
  FSize:= IOStream.Size;
  FDataLen:= DataLen;
  fBufferSize:= BuffSize;
  FRecCount:= BuffSize Div DataLen;
  fBufferSize:= DataLen *FRecCount;
  fNeedFill:= True;
  fEof:= False;
  fFileEof:= False;
  AllocBuffer(fBufferSize);
  fBufferPos:= 0;
end;  { create }

destructor TmSorIO.destroy;
begin
  ReallocMem(fBuffer, 0);
  inherited destroy;
end;  { destroy }

procedure TmSorIO.AllocBuffer(NewValue:Longint);
begin
  fFilledSize:= NewValue;
  ReallocMem(fBuffer, NewValue);
end; { SetBufferSize }

procedure TmSorIO.FillBuffer;
var
  Readed: LongInt;
begin
  Readed:= IOStream.Read(FBuffer^, fBufferSize);
  fFilePos:= fFilePos + Readed;
  if fFilePos = fSize then fFileEof:= True else fFileEof:= False;
  fBufferPos:= 0;
  fFilledSize:= Readed;
  fNeedFill:= False;
end;   { FillBuffer }

function TmSorIO.ReadData:Pointer;
begin
  fEof:= False;
  if fNeedFill then FillBuffer;
  Result:= Pointer(Integer(fBuffer) + fBufferPos);
  inc(fBufferPos, fDataLen);
  if fBufferPos >= fFilledSize then
  begin
    fNeedFill:= True;
    if FFileEof then FEof:= True;
  end;
end;   { ReadData }

procedure TmSorIO.WriteData(Var NewData);
var
  Pos: LongInt;
begin
  if (fBufferPos >= 0) and (Pointer(NewData) <> nil) then
  begin
    Pos := fBufferPos + fDataLen;
    if Pos > 0 then
    begin
      if Pos >= FBufferSize then
        begin
           FlushBuffer;
        end;
      Move(NewData, Pointer(LongInt(fBuffer) + fBufferPos)^, fDataLen);
      inc(fBufferPos, fDataLen);
      inc(fFilePos, fDataLen);
    end;
  end;
end;  { WriteData }

procedure TmSorIO.FlushBuffer;
begin
  IOStream.Write(fBuffer^, fBufferPos);
  fBufferPos:= 0;
end;  { FlushBuffer }

constructor TmMergePart.create(Stream: TFileStream; FilePos, DataLen, Count, aNumber: LongInt);
begin
  fPartStream:= Stream;
  PartFilePos:= FilePos;
  RecsToRead:= Count;
  RecsReaded:= 0;
  FNumber:= aNumber;
  FDataLen:= DataLen;
  FRecCount:= PartBuffSize div DataLen;
  fBufferSize:= DataLen *FRecCount;
  fNeedFill:= True;
  fEof:= False;
  fBufferPos:= 0;
end;  { create }

destructor TmMergePart.destroy;
begin
  ReallocMem(fBuffer, 0);
  inherited destroy;
end;  { destroy }

procedure TmMergePart.AllocBuffer(NewValue:Longint);
begin
  ReallocMem(fBuffer, NewValue);
end; { SetBufferSize }

procedure TmMergePart.FillBuffer;
var
  Readed: LongInt;
begin
  FPartStream.Position:= PartFilePos;
  Readed:= FPartStream.Read(FBuffer^, fBufferSize);
  PartFilePos:= PartFilePos + Readed;
  if Readed = 0 then FEof:= True;
  fBufferPos:= 0;
  fNeedFill:= False;
end;   { FillBuffer }

procedure TmMergePart.Init;
var
  Readed: LongInt;
begin
  AllocBuffer(fBufferSize);
  next;
end;   { Init }

procedure TmMergePart.next;
begin
  fEof:= False;
  if fNeedFill then FillBuffer;
  fData:= Pointer(Integer(fBuffer) + fBufferPos);
  inc(fBufferPos, fDataLen);
  inc(RecsReaded);
  if fBufferPos >= fBufferSize then fNeedFill:= True;
  if RecsReaded = RecsToRead then FEof:= True;
end;   { Read }

constructor TmwSor.Create(RecLen: Integer);
begin
  inherited Create;
  FRecordLen:= RecLen;
end;  { Create }

destructor TmwSor.Destroy;
begin
  inherited Destroy;
end;  { Destroy }

procedure TmwSor.InitMerge;
var
  I: Integer;
begin
  I:= 0;
  while I < FParts.Count do
  begin
    TmMergePart(FParts[I]).Init;
    inc(I);
  end;
  if FParts.Count > 1 then FParts.Sort(StableCompare);
end;  { InitMerge }

{ Similar to the Tree of Looser, but not as effective}
procedure TmwSor.LooserSort;
var
  First, Last, I: Integer;
  Larger: ByteBool;
begin
  if FParts.Count>1 then
  begin
    First:=1;
    Last:=FParts.Count-1;
    while First<=Last do
    begin
      I:=(First+Last)shr 1;
      Case StableCompare(FParts[0], FParts[I])<=0 of
        True:
          begin
            Last:=I-1;
            Larger:=False;
          end;
        False:
          begin
            First:=I+1;
            Larger:=True;
          end;
      end;
    end;
    if I>0 then
      Case Larger of
        True: FParts.Move(0, I);
        False: FParts.Move(0, I-1);
      end;
  end;
end;  { LooserSort }

{Quick and dirty multi merge routine}
procedure TmwSor.Merge;
begin
  WriteStream:= TFileStream.Create(SorFileName, fmOpenWrite);
  Writer:= TmSorIo.create(WriteStream, FRecordLen, WriteBuffSize);
  while FParts.Count > 0 do
  begin
    Writer.WriteData(TmMergePart(FParts[0]).Data^);
    if TmMergePart(FParts[0]).Eof then
    begin
      TmMergePart(FParts[0]).Free;
      FParts.Delete(0);
      if FParts.Count = 0 then
      begin
        FParts.Free;
        break;
      end;
    end else
    begin
      TmMergePart(FParts[0]).Next;
      LooserSort;
    end;
  end;
  Writer.FlushBuffer;
  Writer.Free;
  MergeStream.Free;
  WriteStream.Free;
end;  { Merge }

procedure TmwSor.CalculateBuffers;
var
  Size, PCount, Adjust: LongInt;
  RLen: String;
begin
  Size:=ReadStream.Size;
  if Size mod FRecordLen<>0 then
  begin
    RLen:=IntToStr(FRecordLen);
    raise exception.Create('File can`t be divided through '+RLen);
  end;
  if Size<=40000000 then
  begin
    ReadBuffSize:=Size div 12;
    if ReadBuffSize<325000 then ReadBuffSize:=325000;
    WriteBuffSize:=ReadBuffSize div 5;
  end else;
  if(Size>40000000)and(Size<=100000000)then
  begin
    ReadBuffSize:=Size div 17;
    if ReadBuffSize<4000000 then ReadBuffSize:=4000000;
    WriteBuffSize:=ReadBuffSize div 5;
  end else;
  if(Size>100000000)and(Size<=600000000)then
  begin
    ReadBuffSize:=Size div 25;
    if ReadBuffSize<10000000 then ReadBuffSize:=10000000;
    WriteBuffSize:=ReadBuffSize div 5;
  end else;
  if(Size>600000000)and(Size<=2000000000)then
  begin
    ReadBuffSize:=20000000;
    WriteBuffSize:=ReadBuffSize div 5;
  end;
  Adjust:= ReadBuffSize div fRecordLen;
  ReadBuffSize:= AdJust * fRecordLen;
  PCount:=(Size div ReadBuffSize)+1;
  PartBuffSize:=ReadBuffSize div PCount;
end; { CalculateBuffers }

procedure TmwSor.Start(InFile, OutFile: String;  Compare: TmSorCompare);
var
  aFile, bFile: File;
  K, Readed: Integer;
  WriterPos, SorCount, PartNumber: LongInt;
begin
  TempFileName:= 'SorTemp.mkw';
  SorFileName:= OutFile;
  AssignFile(aFile, TempFileName);
  Rewrite(aFile);
  CloseFile(aFile);
  AssignFile(bFile, SorFileName);
  Rewrite(bFile);
  CloseFile(bFile);
  PartNumber:= 1;
  SorCompare:= Compare;
  SorList:= TmList.Create;
  FParts:= TmList.Create;
  ReadStream:= TFileStream.Create(InFile, fmOpenRead);
  CalculateBuffers;
  MergeStream:= TFileStream.Create(TempFileName, fmOpenReadWrite);
  Reader:= TmSorIo.create(ReadStream, FRecordLen, ReadBuffSize);
  Writer:= TmSorIo.create(MergeStream, FRecordLen, WriteBuffSize);
  while not Reader.Eof do
  begin
    Readed:= 0;
    SorList.Clear;
    while (not Reader.Eof) and (Readed < ReadBuffSize) do
    begin
      SorList.Add(Reader.ReadData);
      inc(Readed, FRecordLen);
    end;
    SorList.Sort(MakeQuickSortStable);
    SorCount:= SorList.Count;
    WriterPos:= Writer.FilePos;
    For K:= 0 to SorList.Count -1 do Writer.WriteData(SorList[K]^);
    FParts.Add(TmMergePart.create(MergeStream, WriterPos, FRecordLen, SorCount, PartNumber));
    inc(PartNumber);
 end;
  Reader.Free;
  Writer.FlushBuffer;
  Writer.Free;
  ReadStream.Free;
  if FParts.Count > 1 then
  begin
    InitMerge;
    Merge;
    DeleteFile(TempFileName);
  end else
  begin
    MergeStream.Free;
    DeleteFile(SorFileName);
    RenameFile(TempFileName, SorFileName);
  end;
end;   { Start }

end.
